from functools import wraps

from ovos_utils.log import log_deprecation

from ovos_workshop.decorators.killable import killable_intent, killable_event
from ovos_workshop.decorators.layers import enables_layer, \
    disables_layer, layer_intent, removes_layer, resets_layers, replaces_layer
from ovos_workshop.decorators.ocp import ocp_play, ocp_pause, ocp_resume, \
    ocp_search, ocp_previous, ocp_featured_media


# TODO: Deprecate unused import retained for backwards-compat.
from ovos_utils import classproperty


def adds_context(context: str, words: str = ''):
    """
    Decorator to add context to the Adapt context manager.

    Args:
        context (str): context Keyword to insert
        words (str): optional string content of Keyword
    """

    def context_add_decorator(func):
        @wraps(func)
        def func_wrapper(*args, **kwargs):
            ret = func(*args, **kwargs)
            args[0].set_context(context, words)
            return ret

        return func_wrapper

    return context_add_decorator


def removes_context(context: str):
    """
    Decorator to remove context from the Adapt context manager.

    Args:
        context (str): Context keyword to remove
    """

    def context_removes_decorator(func):
        @wraps(func)
        def func_wrapper(*args, **kwargs):
            ret = func(*args, **kwargs)
            args[0].remove_context(context)
            return ret

        return func_wrapper

    return context_removes_decorator


def intent_handler(intent_parser: object):
    """
    Decorator for adding a method as an intent handler.
    @param intent_parser: string intent name or adapt.IntentBuilder object
    """

    def real_decorator(func):
        # Store the intent_parser inside the function
        # This will be used later to call register_intent
        if not hasattr(func, 'intents'):
            func.intents = []
        func.intents.append(intent_parser)
        return func

    return real_decorator


def intent_file_handler(intent_file: str):
    """
    Deprecated decorator for adding a method as an intent file handler.
    """

    def real_decorator(func):
        # Store the intent_file inside the function
        # This will be used later to call register_intent_file
        if not hasattr(func, 'intent_files'):
            func.intent_files = []
        func.intent_files.append(intent_file)
        return func

    log_deprecation(f"Use `@intent_handler({intent_file})`", "0.1.0")
    return real_decorator


def resting_screen_handler(name: str):
    """
    Decorator for adding a method as a resting screen handler to optionally
    be shown on screen when device enters idle mode.
    @param name: Name of the restring screen to register
    """

    def real_decorator(func):
        # Store the resting information inside the function
        # This will be used later in register_resting_screen
        if not hasattr(func, 'resting_handler'):
            func.resting_handler = name
        return func

    return real_decorator


def skill_api_method(func: callable):
    """
    Decorator for adding a method to the skill's public api. Methods with this
    decorator will be registered on the messagebus and an api object can be
    created for interaction with the skill.
    @param func: API method to expose
    """
    # tag the method by adding an api_method member to it
    if not hasattr(func, 'api_method') and hasattr(func, '__name__'):
        func.api_method = True
    return func


def converse_handler(func):
    """
    Decorator for aliasing a method as the converse method
    """
    if not hasattr(func, 'converse'):
        func.converse = True
    return func


def conversational_intent(intent_file):
    """Decorator for adding a method as an converse intent handler.
    NOTE: only padatious intents supported, not adapt
    """

    def real_decorator(func):
        # Store the intent_file inside the function
        # This will be used later to train intents
        if not hasattr(func, 'converse_intents'):
            func.converse_intents = []
        func.converse_intents.append(intent_file)
        return func

    return real_decorator


def fallback_handler(priority: int = 50):
    """
    Decorator for adding a fallback intent handler.

    @param priority: Fallback priority (0-100) with lower values having higher
        priority
    """

    def real_decorator(func):
        if not hasattr(func, 'fallback_priority'):
            func.fallback_priority = priority
        return func

    return real_decorator



import json
from os.path import isfile
from threading import Timer
from typing import Optional

import yaml
from json_database import JsonStorageXDG

from ovos_backend_client.api import DeviceApi
from ovos_backend_client.pairing import is_paired, requires_backend
from ovos_backend_client.settings import RemoteSkillSettings, get_display_name
from ovos_bus_client import MessageBusClient
from ovos_bus_client.message import Message, dig_for_message
from ovos_utils.log import LOG


class SkillSettingsManager:
    def __init__(self, skill):
        self.download_timer: Optional[Timer] = None
        self.skill = skill
        self.api = DeviceApi()
        self.remote_settings = \
            RemoteSkillSettings(self.skill_id,
                                settings=dict(self.skill.settings),
                                meta=self.load_meta(), remote_id=self.skill_gid)
        self.register_bus_handlers()

    def start(self):
        self._download()

    def _download(self):
        # If this method is called outside of the timer loop, ensure the
        # existing timer is canceled before starting a new one.
        if self.download_timer:
            self.download_timer.cancel()

        self.download()

        # prepare to download again in 60 seconds
        self.download_timer = Timer(60, self._download)
        self.download_timer.daemon = True
        self.download_timer.start()

    def stop(self):
        # If this method is called outside of the timer loop, ensure the
        # existing timer is canceled
        if self.download_timer:
            self.download_timer.cancel()

    @property
    def bus(self) -> MessageBusClient:
        return self.skill.bus

    @property
    def skill_id(self) -> str:
        return self.skill.skill_id

    @property
    def display_name(self) -> str:
        return get_display_name(self.skill_id)

    @property
    def skill_gid(self) -> str:
        return f"@{self.api.uuid}|{self.skill_id}"

    @property
    def skill_meta(self) -> dict:
        return self.remote_settings.meta

    def register_bus_handlers(self):
        self.skill.add_event('mycroft.skills.settings.update',
                             self.handle_download_remote)  # backwards compat
        self.skill.add_event('mycroft.skills.settings.download',
                             self.handle_download_remote)
        self.skill.add_event('mycroft.skills.settings.upload',
                             self.handle_upload_local)
        self.skill.add_event('mycroft.skills.settings.upload.meta',
                             self.handle_upload_meta)
        self.skill.add_event('mycroft.paired',
                             self.handle_upload_local)

    def load_meta(self) -> dict:
        json_path = f"{self.skill.root_dir}/settingsmeta.json"
        yaml_path = f"{self.skill.root_dir}/settingsmeta.yaml"
        if isfile(yaml_path):
            with open(yaml_path) as meta_file:
                return yaml.safe_load(meta_file)
        elif isfile(json_path):
            with open(json_path) as meta_file:
                return json.load(meta_file)
        return {}

    def save_meta(self, generate: bool = False):
        # unset reload flag to avoid a reload on settingmeta change
        # TODO - support for settingsmeta XDG paths
        reload = self.skill.reload_skill
        self.skill.reload_skill = False

        # generate meta for missing fields
        if generate:
            self.remote_settings.generate_meta()

        # write to disk
        json_path = f"{self.skill.root_dir}/settingsmeta.json"
        yaml_path = f"{self.skill.root_dir}/settingsmeta.yaml"
        if isfile(yaml_path):
            with open(yaml_path) as meta_file:
                yaml.dump(self.remote_settings.meta, meta_file)
        else:
            with open(json_path, "w") as meta_file:
                json.dump(self.remote_settings.meta, meta_file)

        # reset reloading flag
        self.skill.reload_skill = reload

    @requires_backend
    def upload(self, generate: bool = False):
        if not is_paired():
            LOG.error("Device needs to be paired to upload settings")
            return
        self.remote_settings.settings = dict(self.skill.settings)
        if generate:
            self.remote_settings.generate_meta()
        self.remote_settings.upload()

    @requires_backend
    def upload_meta(self, generate: bool = False):
        if not is_paired():
            LOG.error("Device needs to be paired to upload settingsmeta")
            return
        if generate:
            self.remote_settings.settings = dict(self.skill.settings)
            self.remote_settings.generate_meta()
        self.remote_settings.upload_meta()

    @requires_backend
    def download(self):
        if not is_paired():
            LOG.error("Device needs to be paired to download remote settings")
            return
        self.remote_settings.download()
        # we do not update skill object settings directly
        # skill will handle the event and trigger a callback
        if self.skill.settings != self.remote_settings.settings:
            # dig old message to keep context
            msg = dig_for_message() or Message("")
            msg = msg.forward('mycroft.skills.settings.changed')

            msg.data[self.skill_id] = self.remote_settings.settings
            self.bus.emit(msg)

    def handle_upload_meta(self, message: Message):
        skill_id = message.data.get("skill_id")
        if skill_id == self.skill_id:
            self.upload_meta()

    def handle_upload_local(self, message: Message):
        skill_id = message.data.get("skill_id")
        if skill_id == self.skill_id:
            self.upload()

    def handle_download_remote(self, message: Message):
        self.download()


def settings2meta(settings, section_name="Skill Settings"):
    """ generates basic settingsmeta """
    fields = []

    for k, v in settings.items():
        if k.startswith("_"):
            continue
        label = k.replace("-", " ").replace("_", " ").title()
        if isinstance(v, bool):
            fields.append({
                "name": k,
                "type": "checkbox",
                "label": label,
                "value": str(v).lower()
            })
        if isinstance(v, str):
            fields.append({
                "name": k,
                "type": "text",
                "label": label,
                "value": v
            })
        if isinstance(v, int):
            fields.append({
                "name": k,
                "type": "number",
                "label": label,
                "value": str(v)
            })
    return {
        "skillMetadata": {
            "sections": [
                {
                    "name": section_name,
                    "fields": fields
                }
            ]
        }
    }


class PrivateSettings(JsonStorageXDG):
    def __init__(self, skill_id):
        super(PrivateSettings, self).__init__(skill_id)

    @property
    def settingsmeta(self):
        return settings2meta(self, self.name)



try:  # TODO - remove this file in next stable release
    from ovos_utils.ocp import *
except ImportError:
    import inspect
    import mimetypes
    from dataclasses import dataclass
    from enum import IntEnum
    from typing import Optional, Tuple, List, Union

    import orjson
    from ovos_utils.log import LOG, deprecated

    LOG.warning("Please update to ovos-utils~=0.1. Patching missing OCP imports")

    OCP_ID = "ovos.common_play"

    class MatchConfidence(IntEnum):
        EXACT = 95
        VERY_HIGH = 90
        HIGH = 80
        AVERAGE_HIGH = 70
        AVERAGE = 50
        AVERAGE_LOW = 30
        LOW = 15
        VERY_LOW = 1


    class TrackState(IntEnum):
        DISAMBIGUATION = 1  # media result, not queued for playback
        PLAYING_SKILL = 20  # Skill is handling playback internally
        PLAYING_AUDIOSERVICE = 21  ## DEPRECATED - used in ovos 0.0.7
        PLAYING_VIDEO = 22  # Skill forwarded playback to video service
        PLAYING_AUDIO = 23  # Skill forwarded playback to audio service
        PLAYING_MPRIS = 24  # External media player is handling playback
        PLAYING_WEBVIEW = 25  # Media playback handled in browser (eg. javascript)

        QUEUED_SKILL = 30  # Waiting playback to be handled inside skill
        QUEUED_AUDIOSERVICE = 31  ## DEPRECATED - used in ovos 0.0.7
        QUEUED_VIDEO = 32  # Waiting playback in video service
        QUEUED_AUDIO = 33  # Waiting playback in audio service
        QUEUED_WEBVIEW = 34  # Waiting playback in browser service


    class MediaState(IntEnum):
        # https://doc.qt.io/qt-5/qmediaplayer.html#MediaStatus-enum
        # The status of the media cannot be determined.
        UNKNOWN = 0
        # There is no current media. PlayerState == STOPPED
        NO_MEDIA = 1
        # The current media is being loaded. The player may be in any state.
        LOADING_MEDIA = 2
        # The current media has been loaded. PlayerState== STOPPED
        LOADED_MEDIA = 3
        # Playback of the current media has stalled due to
        # insufficient buffering or some other temporary interruption.
        # PlayerState != STOPPED
        STALLED_MEDIA = 4
        # The player is buffering data but has enough data buffered
        # for playback to continue for the immediate future.
        # PlayerState != STOPPED
        BUFFERING_MEDIA = 5
        # The player has fully buffered the current media. PlayerState != STOPPED
        BUFFERED_MEDIA = 6
        # Playback has reached the end of the current media. PlayerState == STOPPED
        END_OF_MEDIA = 7
        # The current media cannot be played. PlayerState == STOPPED
        INVALID_MEDIA = 8


    class PlayerState(IntEnum):
        # https://doc.qt.io/qt-5/qmediaplayer.html#State-enum
        STOPPED = 0
        PLAYING = 1
        PAUSED = 2


    class LoopState(IntEnum):
        NONE = 0
        REPEAT = 1
        REPEAT_TRACK = 2


    class PlaybackType(IntEnum):
        SKILL = 0  # skills handle playback whatever way they see fit,
        # eg spotify / mycroft common play
        VIDEO = 1  # Video results
        AUDIO = 2  # Results should be played audio only
        AUDIO_SERVICE = 3  ## DEPRECATED - used in ovos 0.0.7
        MPRIS = 4  # External MPRIS compliant player
        WEBVIEW = 5  # webview, render a url instead of media player
        UNDEFINED = 100  # data not available, hopefully status will be updated soon..


    class PlaybackMode(IntEnum):
        AUTO = 0  # play each entry as considered appropriate,
        # ie, make it happen the best way possible
        AUDIO_ONLY = 10  # only consider audio entries
        VIDEO_ONLY = 20  # only consider video entries
        FORCE_AUDIO = 30  # cast video to audio unconditionally
        FORCE_AUDIOSERVICE = 40  ## DEPRECATED - used in ovos 0.0.7
        EVENTS_ONLY = 50  # only emit ocp events, do not display or play anything.
        # allows integration with external interfaces


    class MediaType(IntEnum):
        GENERIC = 0  # nothing else matches
        AUDIO = 1  # things like ambient noises
        MUSIC = 2
        VIDEO = 3  # eg, youtube videos
        AUDIOBOOK = 4
        GAME = 5  # because it shares the verb "play", mostly for disambguation
        PODCAST = 6
        RADIO = 7  # live radio
        NEWS = 8  # news reports
        TV = 9  # live tv stream
        MOVIE = 10
        TRAILER = 11
        AUDIO_DESCRIPTION = 12  # narrated movie for the blind
        VISUAL_STORY = 13  # things like animated comic books
        BEHIND_THE_SCENES = 14
        DOCUMENTARY = 15
        RADIO_THEATRE = 16
        SHORT_FILM = 17  # typically movies under 45 min
        SILENT_MOVIE = 18
        VIDEO_EPISODES = 19  # tv series etc
        BLACK_WHITE_MOVIE = 20
        CARTOON = 21
        ANIME = 22
        ASMR = 23

        ADULT = 69  # for content filtering
        HENTAI = 70  # for content filtering
        ADULT_AUDIO = 71  # for content filtering


    def find_mime(uri):
        """ Determine mime type. """
        mime = mimetypes.guess_type(uri)
        if mime:
            return mime
        else:
            return None


    @dataclass
    class MediaEntry:
        uri: str = ""
        title: str = ""
        artist: str = ""
        match_confidence: int = 0  # 0 - 100
        skill_id: str = OCP_ID
        playback: PlaybackType = PlaybackType.UNDEFINED
        status: TrackState = TrackState.DISAMBIGUATION
        media_type: MediaType = MediaType.GENERIC
        length: int = 0  # in seconds
        image: str = ""
        skill_icon: str = ""
        javascript: str = ""  # to execute once webview is loaded

        def update(self, entry: dict, skipkeys: list = None, newonly: bool = False):
            """
            Update this MediaEntry object with keys from the provided entry
            @param entry: dict or MediaEntry object to update this object with
            @param skipkeys: list of keys to not change
            @param newonly: if True, only adds new keys; existing keys are unchanged
            """
            skipkeys = skipkeys or []
            if isinstance(entry, (MediaEntry, PluginStream)):
                entry = entry.as_dict
            entry = entry or {}
            for k, v in entry.items():
                if k not in skipkeys and hasattr(self, k):
                    if newonly and self.__getattribute__(k):
                        # skip, do not replace existing values
                        continue
                    self.__setattr__(k, v)

        @property
        def infocard(self) -> dict:
            """
            Return dict data used for a UI display
            """
            return {
                "duration": self.length,
                "track": self.title,
                "image": self.image,
                "album": self.skill_id,
                "source": self.skill_icon,
                "uri": self.uri
            }

        @property
        def mpris_metadata(self) -> dict:
            """
            Return dict data used by MPRIS
            """
            from dbus_next.service import Variant
            meta = {"xesam:url": Variant('s', self.uri)}
            if self.artist:
                meta['xesam:artist'] = Variant('as', [self.artist])
            if self.title:
                meta['xesam:title'] = Variant('s', self.title)
            if self.image:
                meta['mpris:artUrl'] = Variant('s', self.image)
            if self.length:
                meta['mpris:length'] = Variant('d', self.length)
            return meta

        @property
        def as_dict(self) -> dict:
            """
            Return a dict representation of this MediaEntry
            """
            # orjson handles dataclasses directly
            return orjson.loads(orjson.dumps(self).decode("utf-8"))

        @staticmethod
        def from_dict(track: dict) -> 'MediaEntry':
            if "uri" not in track:
                LOG.error("track dictionary does not contain 'uri', it is not a valid MediaEntry")
                # raise ValueError("track dictionary does not contain 'uri', it is not a valid MediaEntry")
                LOG.warning("DEPRECATED: use dict2entry() for Playlists and PluginStreams,"
                            " MediaEntry.from_dict is only for regular media, will start throwing ValueError in 0.1.0")
                return dict2entry(track)
            kwargs = {k: v for k, v in track.items()
                      if k in inspect.signature(MediaEntry).parameters}
            return MediaEntry(**kwargs)

        @property
        def mimetype(self) -> Optional[Tuple[Optional[str], Optional[str]]]:
            """
            Get the detected mimetype tuple (type, encoding) if it can be determined
            """
            if self.uri:
                return find_mime(self.uri)

        def __eq__(self, other):
            if isinstance(other, MediaEntry):
                other = other.infocard
            # dict comparison
            return other == self.infocard


    @dataclass
    class PluginStream:
        stream: str
        extractor_id: str
        title: str = ""
        artist: str = ""
        match_confidence: int = 0  # 0 - 100
        skill_id: str = OCP_ID
        playback: PlaybackType = PlaybackType.UNDEFINED
        status: TrackState = TrackState.DISAMBIGUATION
        media_type: MediaType = MediaType.GENERIC
        length: int = 0  # in seconds
        image: str = ""
        skill_icon: str = ""

        def extract_uri(self, video=True) -> str:
            from ovos_plugin_manager.ocp import load_stream_extractors
            xtract = load_stream_extractors()
            meta = xtract.extract_stream(f"{self.extractor_id}//{self.stream}",
                                         video=video)
            return meta["uri"]

        def extract_media_entry(self, video=True) -> MediaEntry:
            from ovos_plugin_manager.ocp import load_stream_extractors
            xtract = load_stream_extractors()
            if self.extractor_id not in xtract.supported_seis:
                raise ImportError(f"stream extractor not installed, extractor_id: {self.extractor_id}\navailable plugins: {list(xtract.extractors)}")

            meta = xtract.extract_stream(f"{self.extractor_id}//{self.stream}",
                                         video=video)
            p = meta.get("playback", self.playback)
            if p == PlaybackType.UNDEFINED:
                meta["playback"] = PlaybackType.VIDEO if video else PlaybackType.AUDIO
            for k, v in self.as_dict.items():
                if not meta.get(k):
                    meta[k] = v
            kwargs = {k: v for k, v in meta.items()
                      if k in inspect.signature(MediaEntry).parameters}
            return MediaEntry(**kwargs)

        @property
        def infocard(self) -> dict:
            """
            Return dict data used for a UI display
            (model shared with MediaEntry)
            """
            return {
                "duration": self.length,
                "track": self.title,
                "image": self.image,
                "album": self.skill_id,
                "source": self.skill_icon,
                "uri": f"{self.extractor_id}//{self.stream}"
            }

        @property
        def as_media_entry(self) -> MediaEntry:
            kwargs = {k: v for k, v in self.as_dict.items()
                      if k in inspect.signature(MediaEntry).parameters}
            # TODO - in a couple major versions this should be deprecated
            kwargs["uri"] = f"{self.extractor_id}//{self.stream}"
            return MediaEntry(**kwargs)

        @property
        def as_dict(self) -> dict:
            """
            Return a dict representation of this MediaEntry
            """
            # orjson handles dataclasses directly
            return orjson.loads(orjson.dumps(self).decode("utf-8"))

        @staticmethod
        def from_dict(track: dict) -> 'PluginStream':
            if "extractor_id" not in track:
                raise ValueError("track dictionary does not contain 'extractor_id', it is not a valid PluginStream")
            if "stream" not in track:
                raise ValueError("track dictionary does not contain 'stream', it is not a valid PluginStream")
            kwargs = {k: v for k, v in track.items()
                      if k in inspect.signature(PluginStream).parameters}
            return PluginStream(**kwargs)


    @dataclass
    class Playlist(list):
        title: str = ""
        artist: str = ""
        position: int = 0
        image: str = ""
        match_confidence: int = 0  # 0 - 100
        skill_id: str = OCP_ID
        skill_icon: str = ""
        playback: PlaybackType = PlaybackType.UNDEFINED
        media_type: MediaType = MediaType.GENERIC

        def __init__(self, *args, **kwargs):
            super().__init__()
            for k, v in kwargs.items():
                if hasattr(self, k):
                    try:
                        self.__setattr__(k, v)
                    except AttributeError:
                        continue
            if len(args) == 1 and isinstance(args[0], list):
                args = args[0]
            for e in args:
                self.add_entry(e)

        @property
        def length(self):
            """calc the length value based on all entries"""
            return sum([e.length for e in self.entries])

        @property
        def infocard(self) -> dict:
            """
            Return dict data used for a UI display
            (model shared with MediaEntry)
            """
            return {
                "duration": self.length,
                "track": self.title,
                "image": self.image,
                "album": self.skill_id,
                "source": self.skill_icon,
                "uri": ""
            }

        @staticmethod
        def from_dict(track: dict) -> 'Playlist':
            if "playlist" not in track:
                raise ValueError("track dictionary does not contain 'playlist' entries, it is not a valid Playlist")
            playlist = Playlist(**track)
            for e in track.get("playlist", []):
                playlist.add_entry(e)
            return playlist

        @property
        def as_dict(self) -> dict:
            """
            Return a dict representation of this MediaEntry
            """
            data = {
                "title": self.title,
                "position": self.position,
                "length": self.length,
                "image": self.image,
                "match_confidence": self.match_confidence,
                "skill_id": self.skill_id,
                "skill_icon": self.skill_icon,
                "playlist": [e.as_dict for e in self.entries]
            }
            return data

        @property
        def entries(self) -> List[Union[MediaEntry, PluginStream]]:
            """
            Return a list of MediaEntry objects in the playlist
            """
            entries = []
            for e in self:
                if isinstance(e, dict):
                    e = dict2entry(e)
                if isinstance(e, (MediaEntry, PluginStream)):
                    entries.append(e)
            return entries

        @property
        def current_track(self) -> Optional[Union[MediaEntry, PluginStream]]:
            """
            Return the current MediaEntry or None if the playlist is empty
            """
            if len(self) == 0:
                return None
            self._validate_position()
            track = self[self.position]
            if isinstance(track, dict):
                track = dict2entry(track)
            return track

        @property
        def is_first_track(self) -> bool:
            """
            Return `True` if the current position is the first track or if the
            playlist is empty
            """
            if len(self) == 0:
                return True
            return self.position == 0

        @property
        def is_last_track(self) -> bool:
            """
            Return `True` if the current position is the last track of if the
            playlist is empty
            """
            if len(self) == 0:
                return True
            return self.position == len(self) - 1

        def goto_start(self) -> None:
            """
            Move to the first entry in the playlist
            """
            self.position = 0

        def clear(self) -> None:
            """
            Remove all entries from the Playlist and reset the position
            """
            super().clear()
            self.position = 0

        def sort_by_conf(self):
            """
            Sort the Playlist by `match_confidence` with high confidence first
            """
            self.sort(
                key=lambda k: k.match_confidence if isinstance(k, (MediaEntry, Playlist))
                else k.get("match_confidence", 0), reverse=True)

        def add_entry(self, entry: Union[MediaEntry, PluginStream], index: int = -1) -> None:
            """
            Add an entry at the requested index
            @param entry: MediaEntry to add to playlist
            @param index: index to insert entry at (default -1 to append)
            """
            assert isinstance(index, int)
            if index > len(self):
                raise ValueError(f"Invalid index {index} requested, "
                                 f"playlist only has {len(self)} entries")

            if isinstance(entry, dict):
                entry = dict2entry(entry)

            assert isinstance(entry, (MediaEntry, Playlist, PluginStream))

            if index == -1:
                index = len(self)

            if index < self.position:
                self.set_position(self.position + 1)

            self.insert(index, entry)

        def remove_entry(self, entry: Union[int, dict, MediaEntry, PluginStream]) -> None:
            """
            Remove the requested entry from the playlist or raise a ValueError
            @param entry: index or MediaEntry to remove from the playlist
            """
            if isinstance(entry, int):
                self.pop(entry)
                return
            if isinstance(entry, dict):
                entry = dict2entry(entry)
            assert isinstance(entry, (MediaEntry, PluginStream))
            for idx, e in enumerate(self.entries):
                if e == entry:
                    self.pop(idx)
                    break
            else:
                raise ValueError(f"entry not in playlist: {entry}")

        def replace(self, new_list: List[Union[dict, MediaEntry, PluginStream]]) -> None:
            """
            Replace the contents of this Playlist with new_list
            @param new_list: list of MediaEntry or dict objects to set this list to
            """
            self.clear()
            for e in new_list:
                self.add_entry(e)

        def set_position(self, idx: int):
            """
            Set the position in the playlist to a specific index
            @param idx: Index to set position to
            """
            self.position = idx
            self._validate_position()

        def goto_track(self, track: Union[MediaEntry, dict, PluginStream]) -> None:
            """
            Go to the requested track in the playlist
            @param track: MediaEntry to find and go to in the playlist
            """
            if isinstance(track, dict):
                track = dict2entry(track)

            assert isinstance(track, (MediaEntry, Playlist, PluginStream))

            if isinstance(track, MediaEntry):
                requested_uri = track.uri
            elif isinstance(track, PluginStream):
                requested_uri = track.stream
            else:
                requested_uri = track.title

            for idx, t in enumerate(self):
                if isinstance(t, MediaEntry):
                    pl_entry_uri = t.uri
                elif isinstance(t, PluginStream):
                    pl_entry_uri = t.stream
                else:
                    pl_entry_uri = t.title

                if requested_uri == pl_entry_uri:
                    self.set_position(idx)
                    LOG.debug(f"New playlist position: {self.position}")
                    return
            LOG.error(f"requested track not in the playlist: {track}")

        def next_track(self) -> None:
            """
            Go to the next track in the playlist
            """
            self.set_position(self.position + 1)

        def prev_track(self) -> None:
            """
            Go to the previous track in the playlist
            """
            self.set_position(self.position - 1)

        def _validate_position(self) -> None:
            """
            Make sure the current position is valid; default `position` to 0
            """
            if self.position < 0 or self.position >= len(self):
                LOG.error(f"Playlist pointer is in an invalid position "
                          f"({self.position}! Going to start of playlist")
                self.position = 0

        def __contains__(self, item):
            if isinstance(item, dict):
                item = dict2entry(item)
            for e in self.entries:
                if isinstance(item, PluginStream) and isinstance(e, PluginStream):
                    if e.stream == item.stream and e.extractor_id == item.extractor_id:
                        return True
                elif isinstance(item, MediaEntry) and isinstance(e, MediaEntry):
                    if e.uri == item.uri:
                        return True
            return False


    def dict2entry(track: dict) -> Union[PluginStream, MediaEntry, Playlist]:
        if track.get("playlist"):
            return Playlist.from_dict(track)
        elif track.get("extractor_id"):
            return PluginStream.from_dict(track)
        elif track.get("uri"):
            return MediaEntry.from_dict(track)
        raise ValueError("track dictionary is not a valid MediaEntry, Playlist or PluginStream")



# Copyright 2019 Mycroft AI Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import operator
from typing import Optional, List

from ovos_bus_client import MessageBusClient
from ovos_bus_client.message import Message, dig_for_message
from ovos_config import Configuration
from ovos_utils.events import get_handler_name
from ovos_utils.log import LOG
from ovos_utils.metrics import Stopwatch
from ovos_utils.skills import get_non_properties

from ovos_workshop.decorators.killable import AbortEvent, killable_event
from ovos_workshop.skills.ovos import OVOSSkill


class FallbackSkill(OVOSSkill):
    """
    Fallbacks come into play when no skill matches an Adapt or closely with
    a Padatious intent.  All Fallback skills work together to give them a
    view of the user's utterance.  Fallback handlers are called in an order
    determined the priority provided when the handler is registered.

    ========   ===========================================================
    Priority   Purpose
    ========   ===========================================================
       0-4     High-priority fallbacks before medium-confidence Padatious
      5-89     Medium-priority fallbacks between medium and low Padatious
    90-100     Low-priority fallbacks after all other intent matches

    Handlers with the numerically lowest priority are invoked first.
    Multiple fallbacks can exist at the same priority, but no order is
    guaranteed.

    A Fallback can either observe or consume an utterance. A consumed
    utterance will not be seen by any other Fallback handlers.
    """
    # "skill_id": priority (int)  overrides
    fallback_config = Configuration().get("skills", {}).get("fallbacks", {})

    @classmethod
    def make_intent_failure_handler(cls, bus: MessageBusClient):
        """
        Goes through all fallback handlers until one returns True
        """

        def handler(message):
            # No hard limit to 100, while not officially supported
            # mycroft-lib can handle fallback priorities up to 999
            start, stop = message.data.get('fallback_range', (0, 1000))
            # indicate fallback handling start
            LOG.debug('Checking fallbacks in range '
                      '{} - {}'.format(start, stop))
            bus.emit(message.forward("mycroft.skill.handler.start",
                                     data={'handler': "fallback"}))

            stopwatch = Stopwatch()
            handler_name = None
            with stopwatch:
                sorted_handlers = sorted(cls.fallback_handlers.items(),
                                         key=operator.itemgetter(0))
                handlers = [f[1] for f in sorted_handlers
                            if start <= f[0] < stop]
                for handler in handlers:
                    try:
                        if handler(message):
                            # indicate completion
                            status = True
                            handler_name = get_handler_name(handler)
                            bus.emit(message.forward(
                                'mycroft.skill.handler.complete',
                                data={'handler': "fallback",
                                      "fallback_handler": handler_name}))
                            break
                    except Exception:
                        LOG.exception('Exception in fallback.')
                else:
                    status = False
                    #  indicate completion with exception
                    warning = 'No fallback could handle intent.'
                    bus.emit(message.forward('mycroft.skill.handler.complete',
                                             data={'handler': "fallback",
                                                   'exception': warning}))

            # return if the utterance was handled to the caller
            bus.emit(message.response(data={'handled': status}))

            # Send timing metric
            if message.context.get('ident'):
                ident = message.context['ident']
                cls._report_timing(ident, 'fallback_handler', stopwatch,
                                   {'handler': handler_name})

        return handler

    def __init__(self, bus=None, skill_id="", **kwargs):
        self._fallback_handlers = []
        super().__init__(bus=bus, skill_id=skill_id, **kwargs)

    @property
    def priority(self) -> int:
        """
        Get this skill's minimum priority. Priority is determined as:
            1) Configured fallback skill priority
            2) Highest fallback handler priority
            3) Default `101` (no fallback handlers are registered)
        """
        priority_overrides = self.fallback_config.get("fallback_priorities", {})
        if self.skill_id in priority_overrides:
            return priority_overrides.get(self.skill_id)
        if len(self._fallback_handlers):
            return min([p[0] for p in self._fallback_handlers])
        return 101

    def can_answer(self, utterances: List[str], lang: str) -> bool:
        """
        Check if the skill can answer the particular question. Override this
        method to validate whether a query can possibly be handled. By default,
        assumes a skill can answer if it has any registered handlers
        @param utterances: list of possible transcriptions to parse
        @param lang: BCP-47 language code associated with utterances
        @return: True if skill can handle the query
        """
        return len(self._fallback_handlers) > 0

    def _register_system_event_handlers(self):
        """
        Register messagebus event handlers and emit a message to register this
        fallback skill.
        """
        super()._register_system_event_handlers()
        self.add_event('ovos.skills.fallback.ping', self._handle_fallback_ack, speak_errors=False)
        self.add_event(f"ovos.skills.fallback.{self.skill_id}.request", self._handle_fallback_request,
                       speak_errors=False)

    def _handle_fallback_ack(self, message: Message):
        """
        Inform skills service we can handle fallbacks.
        """
        utts = message.data.get("utterances", [])
        lang = message.data.get("lang")
        self.bus.emit(message.reply("ovos.skills.fallback.pong",
                                    data={"skill_id": self.skill_id,
                                          "can_handle": self.can_answer(utts, lang)},
                                    context={"skill_id": self.skill_id}))

    def _on_timeout(self):
        """_handle_fallback_request timed out and was forcefully killed by ovos-core"""
        message = dig_for_message()
        self.bus.emit(message.forward(f"ovos.skills.fallback.{self.skill_id}.killed",
                                      data={"error": "timed out"}))

    @killable_event("ovos.skills.fallback.force_timeout",
                    callback=_on_timeout, check_skill_id=True)
    def _handle_fallback_request(self, message: Message):
        """
        Handle a fallback request, calling any registered handlers in priority
        order until one is successful. emits a response indicating whether the
        request was handled.
        @param message: `ovos.skills.fallback.<skill_id>.request` message
        """
        # indicate fallback handling start
        self.bus.emit(message.forward(
            f"ovos.skills.fallback.{self.skill_id}.start"))

        handler_name = None

        # each skill can register multiple handlers with different priorities
        sorted_handlers = sorted(self._fallback_handlers,
                                 key=operator.itemgetter(0))
        for prio, handler in sorted_handlers:
            try:
                handler_name = get_handler_name(handler)
                # call handler, conditionally activating the skill
                status = handler(message)
                if status:
                    # indicate completion
                    break
            except AbortEvent:
                LOG.debug(f"fallback handler '{handler_name}' killed because it timed out!")
            except Exception:
                LOG.exception('Exception in fallback.')
        else:
            status = False

        self.bus.emit(message.forward(
            f"ovos.skills.fallback.{self.skill_id}.response",
            data={"result": status, "fallback_handler": handler_name}))

    def register_fallback(self, handler: callable, priority: int):
        """
        Register a fallback handler and add a messagebus handler to call it on
        any fallback request.
        @param handler: Fallback handler
        @param priority: priority of the registered handler
        """
        LOG.info(f"registering fallback handler -> "
                 f"ovos.skills.fallback.{self.skill_id}")
        self._fallback_handlers.append((priority, handler))
        self.bus.on(f"ovos.skills.fallback.{self.skill_id}", handler)
        # register with fallback service
        self.bus.emit(Message("ovos.skills.fallback.register",
                              {"skill_id": self.skill_id,
                               "priority": self.priority}))

    def remove_fallback(self, handler_to_del: Optional[callable] = None) -> bool:
        """
        Remove fallback registration / fallback handler.
        @param handler_to_del: registered callback handler (or wrapped handler)
        @return: True if at least one handler was removed, otherwise False
        """
        found_handler = False
        for i in reversed(range(len(self._fallback_handlers))):
            _, handler = self._fallback_handlers[i]
            if handler_to_del is None or handler == handler_to_del:
                found_handler = True
                del self._fallback_handlers[i]

        if not found_handler:
            LOG.warning('No fallback matching {}'.format(handler_to_del))
        if len(self._fallback_handlers) == 0:
            self.bus.emit(Message("ovos.skills.fallback.deregister",
                                  {"skill_id": self.skill_id}))
        return found_handler

    def default_shutdown(self):
        """
        Remove all registered handlers and perform skill shutdown.
        """
        self.bus.emit(Message("ovos.skills.fallback.deregister",
                              {"skill_id": self.skill_id}))
        self.bus.remove_all_listeners(f"ovos.skills.fallback.{self.skill_id}")
        super().default_shutdown()

    def _register_decorated(self):
        """
        Register all decorated fallback handlers.

        Looks for all functions that have been marked by a decorator
        and read the fallback priority from them. The handlers aren't the
        only decorators used. Skip properties as calling getattr on them
        executes the code which may have unintended side effects.
        """
        super()._register_decorated()
        for attr_name in get_non_properties(self):
            method = getattr(self, attr_name)
            if hasattr(method, 'fallback_priority'):
                self.register_fallback(method, method.fallback_priority)



from ovos_workshop.decorators.layers import IntentLayers
from ovos_workshop.skills.ovos import OVOSSkill
from ovos_workshop.skills.idle_display_skill import IdleDisplaySkill
from ovos_workshop.skills.fallback import FallbackSkill
from ovos_workshop.skills.common_query_skill import CommonQuerySkill
from ovos_workshop.skills.common_play import OVOSCommonPlaybackSkill



import enum

from ovos_config.config import read_mycroft_config, update_mycroft_config


def blacklist_skill(skill, config=None):
    config = config or read_mycroft_config()
    skills_config = config.get("skills", {})
    blacklisted_skills = skills_config.get("blacklisted_skills", [])
    if skill not in blacklisted_skills:
        blacklisted_skills.append(skill)
        conf = {
            "skills": {
                "blacklisted_skills": blacklisted_skills
            }
        }
        update_mycroft_config(conf)
        return True
    return False


def whitelist_skill(skill, config=None):
    config = config or read_mycroft_config()
    skills_config = config.get("skills", {})
    blacklisted_skills = skills_config.get("blacklisted_skills", [])
    if skill in blacklisted_skills:
        blacklisted_skills.pop(skill)
        conf = {
            "skills": {
                "blacklisted_skills": blacklisted_skills
            }
        }
        update_mycroft_config(conf)
        return True
    return False


class ConverseMode(str, enum.Enum):
    """
    Defines a mode for handling `converse` requests.
    ACCEPT ALL - default behavior where all skills may implement `converse`
    WHITELIST - only skills explicitly allowed may implement `converse`
    BLACKLIST - all skills except those disallowed may implement `converse`
    """
    ACCEPT_ALL = "accept_all"  # Default
    WHITELIST = "whitelist"
    BLACKLIST = "blacklist"


class FallbackMode(str, enum.Enum):
    """
    Defines a mode for handling fallbacks (utterances without a matched intent)
    ACCEPT ALL - default behavior where all installed FallbackSkills are used
    WHITELIST - only explicitly allowed FallbackSkills may respond
    BLACKLIST - all FallbackSkills except those disallowed may respond
    """
    ACCEPT_ALL = "accept_all"  # Default
    WHITELIST = "whitelist"
    BLACKLIST = "blacklist"


class ConverseActivationMode(str, enum.Enum):
    """
    Defines a mode for manually activating `converse` handling
    ACCEPT ALL - default behavior where any skill may activate itself
    PRIORITY - a skill may only activate itself if no higher-priority skill is
        currently active
    WHITELIST - only explicitly allowed skills may activate themselves
    BLACKLIST - all skills except those disallowed may activate themselves
    """
    ACCEPT_ALL = "accept_all"  # Default
    PRIORITY = "priority"
    WHITELIST = "whitelist"
    BLACKLIST = "blacklist"
